from typing import Any, Dict, Iterable, List, Tuple , Type, TypeVar, Union

MODEL = TypeVar("MODEL")

class transaction:
    def __init__(self) -> None:
        self.stmt: List[str] = []
    
    def add(self, statement: str) -> None:
        self.stmt.append(statement)
    
    def __iter__(self) -> Iterable:
        return iter(self.stmt)
    
    def __str__(self) -> str:
        return "BEGIN;\n{}\nEND;".format("\n".join(self.stmt))

def create_table(cls: Type[MODEL], force: bool) -> str:
    var: Dict[str, tuple] = cls.__model_fields__
    
    __sql = "--\n-- {}\n--\nCREATE TABLE%s %s(\n\t%s\n);".format(cls.__name__)
    columns = ""

    columns += ",\n\t".join(["%s %s" % (n, k[1]) for n, k in var.items()])

    return __sql % (
        " IF NOT EXISTS" if not force else "", cls.__name__, columns
    )

def drop_table(cls: Type[MODEL], force: bool = False) -> str:
    return "DROP TABLE %s%s;" % ("" if force else "IF EXISTS ", cls.__name__)

class insert_into:
    def __init__(self, modeltype: Type[MODEL]) -> None:
        self.values: List[Tuple[Any]] = []
        self.columns: List[str] = None
        self.modeltype: Type[MODEL] = modeltype
    
    def add(self, args: dict) -> None:
        columns, values = list(zip(*list(args.items())))
        
        if not self.columns:
            self.columns = columns

        self.values.append(values)

    def getvalue(self) -> Union[List, Tuple]:
        return self.values[0] if len(self.values) == 1 else self.values

    def __str__(self) -> str:
        if not self.values:
            raise ValueError(
                'InsertInto: have not any values for creating statement.'
            )
        
        return "INSERT INTO {}{} VALUES({});".format(
            self.modeltype.__name__, '' if not self.columns else '('+','.join(i for i in self.columns)+')',
            ','.join('?' for _ in range(len(self.values[0])))
        )
    
    @classmethod
    def create(cls, model: Type[MODEL], args: dict) -> Tuple[str, tuple]:
        res = cls(model)
        res.add(args)
        return str(res), res.getvalue()

class select:
    def __init__(self, cls: Union[str, Type[MODEL]], kwargs: Dict[str, Any] = None, _limit: int = None, _exclude: bool = False) -> None:
        self.cls = cls if (isinstance(cls, str)) else cls.__name__
        self.kwargs = kwargs
        self._limit = _limit
        self._exclude = _exclude

    def like(self) -> Tuple[str, tuple]:
        __sql = "SELECT * FROM {}".format(self.cls)
        _exclude: str = "{} NOT LIKE ?" if self._exclude else "{} LIKE ?"
        if self.kwargs:
            __sql += " WHERE {}".format(
                " AND ".join(_exclude.format(key) for key in self.kwargs.keys())
            )
        
        if self._limit is not None:
            __sql += " LIMIT %d" % self._limit
    
        return __sql, tuple(self.kwargs.values())

    def condition(self) -> Tuple[str, tuple]:
        __sql = "SELECT * FROM {}".format(self.cls)
        _exclude: str = "{} != ?" if self._exclude else "{} = ?"
        if self.kwargs:
            __sql += " WHERE {}".format(
                " AND ".join(_exclude.format(key) for key in self.kwargs.keys())
            )
        
        if self._limit is not None:
            __sql += " LIMIT %d" % self._limit
    
        return __sql, tuple(self.kwargs.values())
    
    def count(self) ->  Tuple[str, tuple]:
        __sql = "SELECT COUNT(*) FROM {}".format(self.cls)
        _exclude: str = "{} != ?" if self._exclude else "{} = ?"
        if self.kwargs:
            __sql += " WHERE {}".format(
                " AND ".join(_exclude.format(key) for key in self.kwargs.keys())
            )
    
        return __sql, tuple(self.kwargs.values())

def delete(cls: str, kwargs: Dict[str, Any]) ->  Tuple[str, tuple]:
    __sql = "DELETE FROM {}".format(cls)

    __sql += " WHERE {}".format(
        " AND ".join("{} = %s".format(key) for key in kwargs.keys())
    )

    return __sql, tuple(kwargs.values())

def update(cls: str, SET: dict, WHERE: dict) -> Tuple[str, tuple]:
    __sql = "UPDATE {} SET {} WHERE {}".format(cls, ", ".join("{} = ?".format(key) for key in SET.keys()), " AND ".join("{} = ?".format(key) for key in WHERE.keys()))

    return __sql, tuple(SET.values()) + tuple(WHERE.values())
